<!DOCTYPE html>

<html lang="en">
  <head>
    <title>Stats [{package} v{version}]</title>
    <meta charset="UTF-8" />
    <style type="text/css">
      body {
        color: #eee;
        background: #110d0d;
        text-align: center;
        font: 12pt/1.6 Open Sans, Segoe UI, sans-serif;
      }
      #container {
        position: absolute;
        top: 0;
        left: 0;
        margin: 0;
        width: 100%;
        height: 100%;
        display: table;
      }
      #inner {
        width: 100%;
        display: table-cell;
        vertical-align: middle;
      }
      table {
        width: auto;
        margin: auto;
        height: auto;
        background: #222020;
        border: 1px solid #777373;
        border-spacing: 3px;
      }
      th,
      td {
        padding: 6px 18px;
      }
      th {
        font-weight: 700;
        background: linear-gradient(to bottom, #777373, #555151);
      }
      .odd {
        background: #444040;
      }
      .even {
        background: #555151;
      }
      .center {
        text-align: center;
      }
      .right {
        text-align: right;
        font-weight: 600;
      }
    </style>
  </head>

  <body>
    <div id="container">
      <div id="inner">
        <table>
          <tr>
            <th colspan="2">{package} v{version} statistics</th>
          </tr>
          <tr class="odd">
            <td class="right">Open connections</td>
            <td class="center">{opens}</td>
          </tr>

          <tr class="even">
            <td class="right">Bad connections</td>
            <td class="center">{badconns}</td>
          </tr>

          <tr class="odd">
            <td class="right">Denied connections</td>
            <td class="center">{deniedconns}</td>
          </tr>

          <tr class="even">
            <td class="right">Refused (high load)</td>
            <td class="center">{refusedconns}</td>
          </tr>

          <tr class="odd">
            <td class="right">Total requests</td>
            <td class="center">{reqs}</td>
          </tr>
        </table>
      </div>
    </div>
  </body>
</html>
